/*
 *  Copyright (C) 2006 Ludovic Jacomme (ludovic.jacomme@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>

#include <pspctrl.h>
#include <pspkernel.h>
#include <pspdebug.h>

#include "psp_kbd.h"
#include "psp_pg.h"

#include "hp48.h"

# if 0 //LUDO:
# define DEBUG_BUTTONS
# else
# endif
# define KBD_MIN_ANALOG_TIME 150000

 static SceCtrlData  loc_button_data;
 static unsigned int loc_last_analog_time = 0;
 static char         loc_button_press[ KBD_MAX_BUTTONS ]; 
 static char         loc_button_release[ KBD_MAX_BUTTONS ]; 
 static unsigned int loc_button_mask[ KBD_MAX_BUTTONS ] =
 {
   PSP_CTRL_UP         , /*  KBD_UP         */
   PSP_CTRL_RIGHT      , /*  KBD_RIGHT      */
   PSP_CTRL_DOWN       , /*  KBD_DOWN       */
   PSP_CTRL_LEFT       , /*  KBD_LEFT       */
   PSP_CTRL_TRIANGLE   , /*  KBD_TRIANGLE   */
   PSP_CTRL_CIRCLE     , /*  KBD_CIRCLE     */
   PSP_CTRL_CROSS      , /*  KBD_CROSS      */
   PSP_CTRL_SQUARE     , /*  KBD_SQUARE     */
   PSP_CTRL_SELECT     , /*  KBD_SELECT     */
   PSP_CTRL_START      , /*  KBD_START      */
   PSP_CTRL_HOME       , /*  KBD_HOME       */
   PSP_CTRL_HOLD       , /*  KBD_HOLD       */
   PSP_CTRL_LTRIGGER   , /*  KBD_LTRIGGER   */
   PSP_CTRL_RTRIGGER   , /*  KBD_RTRIGGER   */
 };

#define HP48_BUTTON_A        0
#define HP48_BUTTON_B        1
#define HP48_BUTTON_C        2
#define HP48_BUTTON_D        3
#define HP48_BUTTON_E        4
#define HP48_BUTTON_F        5

#define HP48_BUTTON_MTH        6
#define HP48_BUTTON_PRG        7
#define HP48_BUTTON_CST        8
#define HP48_BUTTON_VAR        9
#define HP48_BUTTON_UP        10
#define HP48_BUTTON_NXT        11

#define HP48_BUTTON_COLON        12
#define HP48_BUTTON_STO        13
#define HP48_BUTTON_EVAL        14
#define HP48_BUTTON_LEFT        15
#define HP48_BUTTON_DOWN        16
#define HP48_BUTTON_RIGHT        17

#define HP48_BUTTON_SIN        18
#define HP48_BUTTON_COS        19
#define HP48_BUTTON_TAN        20
#define HP48_BUTTON_SQRT        21
#define HP48_BUTTON_POWER        22
#define HP48_BUTTON_INV        23

#define HP48_BUTTON_ENTER        24
#define HP48_BUTTON_NEG        25
#define HP48_BUTTON_EEX        26
#define HP48_BUTTON_DEL        27
#define HP48_BUTTON_BS        28

#define HP48_BUTTON_ALPHA        29
#define HP48_BUTTON_7        30
#define HP48_BUTTON_8        31
#define HP48_BUTTON_9        32
#define HP48_BUTTON_DIV        33

#define HP48_BUTTON_SHL        34
#define HP48_BUTTON_4        35
#define HP48_BUTTON_5        36
#define HP48_BUTTON_6        37
#define HP48_BUTTON_MUL        38

#define HP48_BUTTON_SHR        39
#define HP48_BUTTON_1        40
#define HP48_BUTTON_2        41
#define HP48_BUTTON_3        42
#define HP48_BUTTON_MINUS        43

#define HP48_BUTTON_ON        44
#define HP48_BUTTON_0        45
#define HP48_BUTTON_PERIOD        46
#define HP48_BUTTON_SPC        47
#define HP48_BUTTON_PLUS        48

#define HP48_LAST_BUTTON        48


  typedef struct hp48_button_t {
    char                *name;
    short                pressed;
    short                key_down;
    int                  code;
    int                  x, y;
    unsigned int         w, h;
  } hp48_button_t;

  static int hp48_cursor_button = HP48_BUTTON_ON;
  static int hp48_cursor_button_on = 0;

  static hp48_button_t hp48_buttons_gx[] = 
  {
    { "A",       0, 0, 0x14, 301,   6, 18, 32 }, 
    { "B",       0, 0, 0x84, 301,  51, 18, 32 }, 
    { "C",       0, 0, 0x83, 301,  96, 18, 32 }, 
    { "D",       0, 0, 0x82, 301, 141, 18, 32 }, 
    { "E",       0, 0, 0x81, 301, 186, 18, 32 }, 
    { "F",       0, 0, 0x80, 301, 231, 18, 32 }, 

    { "MTH",     0, 0, 0x24, 262,   6, 19, 32 }, 
    { "PRG",     0, 0, 0x74, 262,  51, 19, 32 }, 
    { "CST",     0, 0, 0x73, 262,  96, 19, 32 }, 
    { "VAR",     0, 0, 0x72, 262, 141, 19, 32 }, 
    { "UP",      0, 0, 0x71, 262, 186, 19, 32 }, 
    { "NXT",     0, 0, 0x70, 262, 231, 19, 32 }, 

    { "COLON",   0, 0, 0x04, 225,   6, 19, 32 }, 
    { "STO",     0, 0, 0x64, 225,  51, 19, 32 }, 
    { "EVAL",    0, 0, 0x63, 225,  96, 19, 32 }, 
    { "LEFT",    0, 0, 0x62, 225, 141, 19, 32 }, 
    { "DOWN",    0, 0, 0x61, 225, 186, 19, 32 }, 
    { "RIGHT",   0, 0, 0x60, 225, 231, 19, 32 }, 

    { "SIN",     0, 0, 0x34, 188,   6, 19, 32 }, 
    { "COS",     0, 0, 0x54, 188,  51, 19, 32 }, 
    { "TAN",     0, 0, 0x53, 188,  96, 19, 32 }, 
    { "SQRT",    0, 0, 0x52, 188, 141, 19, 32 }, 
    { "POWER",   0, 0, 0x51, 188, 186, 19, 32 }, 
    { "INV",     0, 0, 0x50, 188, 231, 19, 32 }, 

    { "ENTER",   0, 0, 0x44, 151,   6, 19, 77 }, 
    { "NEG",     0, 0, 0x43, 151,  96, 19, 32 }, 
    { "EEX",     0, 0, 0x42, 151, 141, 19, 32 }, 
    { "DEL",     0, 0, 0x41, 151, 186, 19, 32 }, 
    { "BS",      0, 0, 0x40, 151, 231, 19, 32 }, 

    { "ALPHA",   0, 0, 0x35, 114,   6, 18, 32 }, 
    { "7",       0, 0, 0x33, 114,  60, 18, 41 }, 
    { "8",       0, 0, 0x32, 114, 114, 18, 41 }, 
    { "9",       0, 0, 0x31, 114, 168, 18, 41 }, 
    { "DIV",     0, 0, 0x30, 114, 222, 18, 41 }, 

    { "SHL",     0, 0, 0x25,  77,   6, 18, 32 }, 
    { "4",       0, 0, 0x23,  77,  60, 18, 41 }, 
    { "5",       0, 0, 0x22,  77, 114, 18, 41 }, 
    { "6",       0, 0, 0x21,  77, 168, 18, 41 }, 
    { "MUL",     0, 0, 0x20,  77, 222, 18, 41 }, 

    { "SHR",     0, 0, 0x15,  39,   6, 19, 32 }, 
    { "1",       0, 0, 0x13,  39,  60, 19, 41 }, 
    { "2",       0, 0, 0x12,  39, 114, 19, 41 }, 
    { "3",       0, 0, 0x11,  39, 168, 19, 41 }, 
    { "MINUS",   0, 0, 0x10,  39, 222, 19, 41 }, 

    { "ON",      0, 0, 0x8000, 2,   6, 19, 32 }, 
    { "0" ,      0, 0, 0x03,   2,  60, 19, 41 }, 
    { "PERIOD",  0, 0, 0x02,   2, 114, 19, 41 }, 
    { "SPC",     0, 0, 0x01,   2, 168, 19, 41 }, 
    { "PLUS",    0, 0, 0x00,   2, 222, 19, 41 }, 

    { 0 }
  };

int
key_event(int b, int button_pressed)
{
  int code;
  int i, r, c;
  int all_up;

  code = hp48_buttons_gx[b].code;
  if (button_pressed) {
    hp48_buttons_gx[b].key_down = 1;
    hp48_buttons_gx[b].pressed = 1;
    //DrawButton(b);
    if (code == 0x8000) {
      for (i = 0; i < 9; i++)
        saturn.keybuf.rows[i] |= 0x8000;
      do_kbd_int();
    } else {
      r = code >> 4;
      c = 1 << (code & 0xf);
      if ((saturn.keybuf.rows[r] & c) == 0) {
        if (saturn.kbd_ien) {
          do_kbd_int();
        }
        saturn.keybuf.rows[r] |= c;
      }
    }
#ifdef DEBUG_BUTTONS
    pspDebugPrintf( "Button pressed  %d (%s) PC=%x\n",
            hp48_buttons_gx[b].code, hp48_buttons_gx[b].name, saturn.PC);
#endif
  } else {
    hp48_buttons_gx[b].key_down = 0;
    all_up = 1;
    for (i = HP48_BUTTON_A; i <= HP48_LAST_BUTTON; i++) {
      if (hp48_buttons_gx[i].pressed) {
        if (hp48_buttons_gx[i].key_down) {
          all_up = 0;
        } else {
          hp48_buttons_gx[i].pressed = 0;
          //DrawButton(i);
          if (code == 0x8000) {
            for (i = 0; i < 9; i++)
              saturn.keybuf.rows[i] &= ~0x8000;
          } else {
            r = code >> 4;
            c = 1 << (code & 0xf);
            saturn.keybuf.rows[r] &= ~c;
          }
        }
      }
    }
    if (all_up) {
      memset(&saturn.keybuf, 0, sizeof(saturn.keybuf));
    }
#ifdef DEBUG_BUTTONS
    pspDebugPrintf( "Button released %d (%s) PC=%x\n",
            hp48_buttons_gx[b].code, hp48_buttons_gx[b].name, saturn.PC);
#endif
  }
  return 0;
}

int
hp48_decode_key(int psp_b, int button_pressed)
{
  int wake = 0;
  switch ( psp_b )  {
# ifndef KBD_REVERSE_LAYOUT
    case KBD_LEFT     : wake = 1;
                        key_event( HP48_BUTTON_DOWN, button_pressed);
    break;
    case KBD_RIGHT    : wake = 1;
                        key_event( HP48_BUTTON_UP, button_pressed);
    break;
    case KBD_UP       : wake = 1;
                        key_event( HP48_BUTTON_LEFT, button_pressed);
    break;
    case KBD_DOWN     : wake = 1;
                        key_event( HP48_BUTTON_RIGHT, button_pressed);
    break;
    case KBD_CROSS    : wake = 1;
                        key_event( HP48_BUTTON_ALPHA, button_pressed);
    break;
    case KBD_SQUARE   : wake = 1;
                        key_event( HP48_BUTTON_ENTER, button_pressed);
    break;
    case KBD_CIRCLE   : wake = 1;
                        key_event( hp48_cursor_button, button_pressed);
    break;
    case KBD_TRIANGLE : wake = 1;
                        key_event( HP48_BUTTON_BS, button_pressed);
    break;
# else
    case KBD_SQUARE   : wake = 1;
                        key_event( HP48_BUTTON_DOWN, button_pressed);
    break;
    case KBD_CIRCLE   : wake = 1;
                        key_event( HP48_BUTTON_UP, button_pressed);
    break;
    case KBD_TRIANGLE : wake = 1;
                        key_event( HP48_BUTTON_LEFT, button_pressed);
    break;
    case KBD_CROSS    : wake = 1;
                        key_event( HP48_BUTTON_RIGHT, button_pressed);
    break;
    case KBD_DOWN     : wake = 1;
                        key_event( HP48_BUTTON_ALPHA, button_pressed);
    break;
    case KBD_LEFT     : wake = 1;
                        key_event( HP48_BUTTON_ENTER, button_pressed);
    break;
    case KBD_RIGHT    : wake = 1;
                        key_event( hp48_cursor_button, button_pressed);
    break;
    case KBD_UP       : wake = 1;
                        key_event( HP48_BUTTON_BS, button_pressed);
    break;

# endif
    case KBD_START    : wake = 1;
                        key_event( HP48_BUTTON_ON, button_pressed);
    break;
    case KBD_SELECT   : wake = 1;
                        key_event( HP48_BUTTON_NXT, button_pressed);
    break;
    case KBD_LTRIGGER : wake = 1;
                        key_event( HP48_BUTTON_SHR, button_pressed);
    break;
    case KBD_RTRIGGER : wake = 1;
                        key_event( HP48_BUTTON_SHL, button_pressed);
    break;
  }

  return wake;
}

static void
hp48_show_cursor_button(void)
{
  int b = hp48_cursor_button;
  if (! hp48_cursor_button_on) {
    hp48_cursor_button_on = 1;
    pgFillRectangle( hp48_buttons_gx[b].x, 
                     hp48_buttons_gx[b].y,
                     hp48_buttons_gx[b].w,
                     hp48_buttons_gx[b].h, PG_RGB(0xFF,0xFF,0xFF), PG_XOR);
  }
}

static void
hp48_hide_cursor_button(void)
{
  int b = hp48_cursor_button;
  if (hp48_cursor_button_on) {
    hp48_cursor_button_on = 0;
    pgFillRectangle( hp48_buttons_gx[b].x, 
                     hp48_buttons_gx[b].y,
                     hp48_buttons_gx[b].w,
                     hp48_buttons_gx[b].h, PG_RGB(0xFF,0xFF,0xFF), PG_XOR);
  }
}

static void
hp48_move_cursor_button_right(void)
{
  if ((hp48_cursor_button >= HP48_BUTTON_ON) && 
      (hp48_cursor_button <  HP48_BUTTON_PLUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHR  ) && 
      (hp48_cursor_button <  HP48_BUTTON_MINUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHL) && 
      (hp48_cursor_button <  HP48_BUTTON_MUL)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ALPHA) && 
      (hp48_cursor_button <  HP48_BUTTON_DIV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ENTER) && 
      (hp48_cursor_button <  HP48_BUTTON_BS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SIN) && 
      (hp48_cursor_button <  HP48_BUTTON_INV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_COLON) && 
      (hp48_cursor_button <  HP48_BUTTON_RIGHT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_MTH) && 
      (hp48_cursor_button <  HP48_BUTTON_NXT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_A) && 
      (hp48_cursor_button <  HP48_BUTTON_F)) {
    hp48_hide_cursor_button();
    hp48_cursor_button++;
  }
}

static void
hp48_move_cursor_button_left(void)
{
  if ((hp48_cursor_button >  HP48_BUTTON_ON) && 
      (hp48_cursor_button <= HP48_BUTTON_PLUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_SHR  ) && 
      (hp48_cursor_button <= HP48_BUTTON_MINUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_SHL) && 
      (hp48_cursor_button <= HP48_BUTTON_MUL)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_ALPHA) && 
      (hp48_cursor_button <= HP48_BUTTON_DIV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_ENTER) && 
      (hp48_cursor_button <= HP48_BUTTON_BS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_SIN) && 
      (hp48_cursor_button <= HP48_BUTTON_INV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_COLON) && 
      (hp48_cursor_button <= HP48_BUTTON_RIGHT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_MTH) && 
      (hp48_cursor_button <= HP48_BUTTON_NXT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  } else
  if ((hp48_cursor_button >  HP48_BUTTON_A) && 
      (hp48_cursor_button <= HP48_BUTTON_F)) {
    hp48_hide_cursor_button();
    hp48_cursor_button--;
  }
}

static void
hp48_move_cursor_button_up(void)
{
  if ((hp48_cursor_button >= HP48_BUTTON_ON) && 
      (hp48_cursor_button <= HP48_BUTTON_PLUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_ON) + HP48_BUTTON_SHR;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHR  ) && 
      (hp48_cursor_button <= HP48_BUTTON_MINUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_SHR) + HP48_BUTTON_SHL;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHL) && 
      (hp48_cursor_button <= HP48_BUTTON_MUL)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_SHL) + HP48_BUTTON_ALPHA;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ALPHA) && 
      (hp48_cursor_button <= HP48_BUTTON_DIV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_ALPHA) + HP48_BUTTON_ENTER;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ENTER) && 
      (hp48_cursor_button <= HP48_BUTTON_BS)) {
    hp48_hide_cursor_button();
    if (hp48_cursor_button == HP48_BUTTON_ENTER) {
      hp48_cursor_button = HP48_BUTTON_SIN;
    } else {
      hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_ENTER) + HP48_BUTTON_COS;
    }
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SIN) && 
      (hp48_cursor_button <= HP48_BUTTON_INV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_SIN) + HP48_BUTTON_COLON;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_COLON) && 
      (hp48_cursor_button <= HP48_BUTTON_RIGHT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_COLON) + HP48_BUTTON_MTH;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_MTH) && 
      (hp48_cursor_button <= HP48_BUTTON_NXT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_MTH) + HP48_BUTTON_A;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_A) && 
      (hp48_cursor_button <= HP48_BUTTON_F)) {
    /* Nothing to do */
  }
}

static void
hp48_move_cursor_button_down(void)
{
  if ((hp48_cursor_button >= HP48_BUTTON_ON) && 
      (hp48_cursor_button <= HP48_BUTTON_PLUS)) {
     /* Nothing to do */
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHR  ) && 
      (hp48_cursor_button <= HP48_BUTTON_MINUS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_SHR) + HP48_BUTTON_ON;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SHL) && 
      (hp48_cursor_button <= HP48_BUTTON_MUL)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_SHL) + HP48_BUTTON_SHR;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ALPHA) && 
      (hp48_cursor_button <= HP48_BUTTON_DIV)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_ALPHA) + HP48_BUTTON_SHL;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_ENTER) && 
      (hp48_cursor_button <= HP48_BUTTON_BS)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_ENTER) + HP48_BUTTON_ALPHA;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_SIN) && 
      (hp48_cursor_button <= HP48_BUTTON_INV)) {
    hp48_hide_cursor_button();
    if ((hp48_cursor_button == HP48_BUTTON_SIN) || (hp48_cursor_button == HP48_BUTTON_COS)) {
      hp48_cursor_button = HP48_BUTTON_ENTER;
    } else {
      hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_COS) + HP48_BUTTON_ENTER;
    }
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_COLON) && 
      (hp48_cursor_button <= HP48_BUTTON_RIGHT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_COLON) + HP48_BUTTON_SIN;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_MTH) && 
      (hp48_cursor_button <= HP48_BUTTON_NXT)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_MTH) + HP48_BUTTON_COLON;
  } else
  if ((hp48_cursor_button >= HP48_BUTTON_A) && 
      (hp48_cursor_button <= HP48_BUTTON_F)) {
    hp48_hide_cursor_button();
    hp48_cursor_button = (hp48_cursor_button - HP48_BUTTON_A) + HP48_BUTTON_MTH;
  }
}

int 
kbd_get_analog_direction(int Analog_xy)
{
  if (Analog_xy <=  50) return -1; /* Left */
  else 
  if (Analog_xy >= 200) return  1; /* Right */
  return 0;
}

int
kbd_scan_keyboard(void)
{
  SceCtrlData c;
  int         event;
  int         b;

  int new_Lx;
  int new_Ly;
  int old_Lx;
  int old_Ly;

# if 0 
  pspDebugPrintf("TimeStamp=%d Buttons=%x Lx=%d Ly=%d\n",
      ctrl.TimeStamp, ctrl.Buttons, ctrl.Lx, ctrl.Ly);
# endif
  event = 0;
  sceCtrlPeekBufferPositive( &c, 1 );
  // sceCtrlReadBufferPositive(&c, 1);

  if ((c.Buttons & (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) ==
       (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) {
    /* Exit ! */
    psp_exit(0);
  } else if (c.Buttons & (PSP_CTRL_START|PSP_CTRL_SELECT) == (PSP_CTRL_START|PSP_CTRL_SELECT)) {
    /* Save ram */
    save_emulator_stuff();
  }

  hp48_show_cursor_button();
  
  /* Check Analog Device */
  old_Lx = kbd_get_analog_direction(loc_button_data.Lx);
  old_Ly = kbd_get_analog_direction(loc_button_data.Ly);
  new_Lx = kbd_get_analog_direction(c.Lx);
  new_Ly = kbd_get_analog_direction(c.Ly);

  if ((c.TimeStamp - loc_last_analog_time) > KBD_MIN_ANALOG_TIME) {
    loc_last_analog_time = c.TimeStamp;
    /* Analog device has moved */
    if (new_Lx > 0) {
      hp48_move_cursor_button_up();
    } else 
    if (new_Lx < 0) {
      hp48_move_cursor_button_down();
    }

    if (new_Ly > 0) {
      hp48_move_cursor_button_right();
    } else 
    if (new_Ly < 0) {
      hp48_move_cursor_button_left();
    }
  }

  for (b = 0; b < KBD_MAX_BUTTONS; b++) {
    if (c.Buttons & loc_button_mask[b]) {
      if (!(loc_button_data.Buttons & loc_button_mask[b])) {
        loc_button_press[b] = 1;
        event = 1;
      }
    } else {
      if (loc_button_data.Buttons & loc_button_mask[b]) {
        loc_button_release[b] = 1;
        event = 1;
      }
    }
  }
  memcpy(&loc_button_data,&c,sizeof(SceCtrlData));

  return event;
}

void
kbd_wait_start(void)
{
  while (1)
  {
    SceCtrlData c;
    sceCtrlReadBufferPositive(&c, 1);
    if (c.Buttons & PSP_CTRL_START) break;
  }
}

/* Event */
int
psp_treat_event(void)
{
# if 0 //LUDO: DEBUG
  static int  first_time = 1;
  static long loc_iter = 150000;
  static int  key_step = 1;
  static int  key_num  = HP48_BUTTON_0;

  if (first_time) {
    first_time = 0;
    update_display();
    redraw_annunc();
    return 0;
  }
  if (--loc_iter <= 0) {
    key_event( key_num, key_step);
    if (key_step) loc_iter = 30000;
    else {
      loc_iter = 40000; key_num++; 
      if (key_num >= HP48_LAST_BUTTON) key_num = HP48_BUTTON_0; 
    }
    key_step = ! key_step;
    return 1;
  }
  {
    SceCtrlData c;
    sceCtrlPeekBufferPositive( &c, 1 );

    if ((c.Buttons & (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) ==
         (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) {
      /* Exit ! */
      psp_exit(0);
    }
  }
  return 1;
# else
  static char first_time = 1;
  static int release_pending = 0;
  static int release_button  = 0;
  //static int last_release_time = 0;

  int b;
  int wake = 0;

  if (first_time) {
    first_time = 0;

    update_display();
    redraw_annunc();

    release_pending = 0;
    release_button  = HP48_BUTTON_ON;

    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      loc_button_release[b] = 0;
      loc_button_press[b] = 0;
    }
    sceCtrlPeekBufferPositive(&loc_button_data, 1);
    return 0;
  }

  if (release_pending)
  {
    release_pending = 0;
    wake = hp48_decode_key(release_button,0);
    update_display();
    redraw_annunc();

    return wake;
  }

  if (kbd_scan_keyboard()) {

    /* check release event */
    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      if (loc_button_release[b]) {
        loc_button_release[b] = 0;
        release_button        = b;
        release_pending       = 1;
        break;
      } 
    }
    /* check press event */
    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      if (loc_button_press[b]) {
        loc_button_press[b] = 0;
        release_pending     = 0;
        wake = hp48_decode_key(b, 1);
        break;
      }
    }
  }
  return wake;
# endif
}
